non_metric_results = {
    "comment",
    "cli_args/comment",
    "run_id",
    "experiment_name",
    "experiment_group",
    "experiment_key",
    "trainable_name",
    "env_runners/environments/seeds",
    "env_runners/environments/seeds/seed_sequence",
    "evaluation/env_runners/environments/seeds",
    "evaluation/env_runners/environments/seeds/seed_sequence",
    # New log style
    "training/environments/seeds",
    "training/environments/seeds/seed_sequence",
    "evaluation/environments/seeds",
    "evaluation/environments/seeds/seed_sequence",
}

exclude_results = {
    "cli_args/test",
    "cli_args/num_jobs",
    "node_ip",
}
